# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Environment wrappers."""
import functools
from brax.v1.envs import env as brax_env
from jax import numpy as jnp


class ExtraStepArgsWrapper(brax_env.Wrapper):
  """Add optional kwargs to step() in Brax env."""

  def step(self, state: brax_env.State, action: jnp.ndarray, **kwargs):
    self.unwrapped.step = functools.partial(self.unwrapped.step, **kwargs)
    return self.env.step(state, action)
