# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry of BIG-Gym environments and components."""

# keep alphabetical ordering
REGISTRIES = [
    'jump',
    'proant',
    'ant',
]


def get_comp_name(registry_name: str, comp_name: str):
  return f'{registry_name}__{comp_name}'


def get_env_name(registry_name: str, env_name: str):
  return f'{registry_name}__{env_name}'
